# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import unittest
from unittest.mock import patch

import pandas as pd
from pandas.testing import assert_frame_equal
import pkg_resources
from datetime import datetime

from Xpress_parse.write import *


class TestWrite(unittest.TestCase):

    def setUp(self) -> None:
        self.root = pkg_resources.resource_filename('Xpress_parse', 'test')
        self.tsv = '%s/input.tsv' % self.root
        self.args = {'input_fps': ('/a/b.txt',), 'output_fp': '/some/path.tsv'}
        self.record_pd = pd.DataFrame(
            [['1', 'f1', 'you'],
             ['2', 'f2', 'you'],
             ['3', 'f3', 'you'],
             ['4', 'f4', 'you'],
             ['5', 'f5', 'you']],
            columns=['AB', 'file', 'PERSON'],
            index=[0, 1, 2, 3, 4])
        self.output_fp = '%s/out.tsv' % pkg_resources.resource_filename(
            'Xpress_parse', 'test')

    @patch('builtins.print')
    def test_write_no_selection(self, mock_print):
        selected = {'record#0': (False, None),
                    'record#1': (False, None),
                    'record#2': (False, None),
                    'record#3': (False, None),
                    'record#4': (False, None)}

        with self.assertRaises(SystemExit) as cm:
            write_selection(self.output_fp, self.record_pd, selected)
        self.assertEqual(cm.exception.code, 0)

        self.assertFalse(os.path.isfile(self.output_fp))

        exp = 'Nothing selected'
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(exp, print_out)

    @patch('builtins.print')
    def test_write_selection(self, mock_print):
        selected = {'record#0': (True, ['no_label']),
                    'record#1': (False, None),
                    'record#2': (False, None),
                    'record#3': (False, None),
                    'record#4': (False, None)}
        out_pd = pd.DataFrame([['1', 'f1', 'you', 'no_label']],
                              columns=['AB', 'file', 'PERSON', 'labels'],
                              index=[0])
        write_selection(self.output_fp, self.record_pd, selected)
        print_out = mock_print.call_args_list[0][0][0]

        self.assertTrue(os.path.isfile(self.output_fp))

        obs_pd = pd.read_csv(self.output_fp, sep='\t', dtype=str)
        assert_frame_equal(out_pd, obs_pd)

        exp = 'Written: %s' % self.output_fp
        self.assertEqual(exp, print_out)

    @patch('builtins.print')
    def test_write_selection_filled(self, mock_print):
        selected = {'record#0': (False, None),
                    'record#1': (True, ['no_label']),
                    'record#2': (True, ['a']),
                    'record#3': (True, ['a', 'b']),
                    'record#4': (False, None)}
        out_pd = pd.DataFrame([['2', 'f2', 'you', 'no_label'],
                               ['3', 'f3', 'you', 'a'],
                               ['4', 'f4', 'you', 'a__b']],
                              columns=['AB', 'file', 'PERSON', 'labels'],
                              index=[0, 1, 2])
        write_selection(self.output_fp, self.record_pd, selected)
        print_out = mock_print.call_args_list[0][0][0]

        self.assertTrue(os.path.isfile(self.output_fp))

        obs_pd = pd.read_csv(self.output_fp, sep='\t', dtype=str)
        assert_frame_equal(out_pd, obs_pd)

        exp = 'Written: %s' % self.output_fp
        self.assertEqual(exp, print_out)

    def test_get_output_fp(self):
        obs = get_output_fp(self.args, 'whatever', '')
        self.assertEqual('/some/path.tsv', obs)

    def test_get_output_fp_exist(self):
        self.args['output_fp'] = None
        cur_date = datetime.now().strftime("%Y%m%d")
        obs = get_output_fp(self.args, '', cur_date)
        exp = '/a/b_xpressed_%s.tsv' % cur_date
        self.assertEqual(exp, obs)

    def test_get_output_fp_user(self):
        self.args['output_fp'] = None
        cur_date = datetime.now().strftime("%Y%m%d")
        obs = get_output_fp(self.args, 'USER', cur_date)
        exp = '/a/b_xpressed_%s_USER.tsv' % cur_date
        self.assertEqual(exp, obs)

    def tearDown(self) -> None:
        if os.path.isfile(self.output_fp):
            os.remove(self.output_fp)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the MIT License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from Xpress_parse.xpress import xpress
from Xpress_parse import __version__


@click.command()
@click.option(
    "-i", "--i-records", multiple=True, help="Path(s) to WoS search result(s)")
@click.option(
    "-o", "--o-output", default=None, type=str,
    help="Selected records (default: <input>_xpressed_YYYYMMDD.tsv)")
@click.option(
    "-u", "--p-user", multiple=True, default=None, help="User name(s)")
@click.option(
    "-l", "--p-labels", default=None,
    help="Path to file with 2 tab-separated columns: keywords and labels")
@click.version_option(__version__, prog_name="Xpress_parse")


def standalone_xpress(
        i_records,
        o_output,
        p_user,
        p_labels,
):

    xpress(
        input_fps=i_records,
        output_fp=o_output,
        user=p_user,
        labels_fp=p_labels,
    )


if __name__ == "__main__":
    standalone_xpress()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from datetime import datetime
from Xpress_parse.records import get_records
from Xpress_parse.users import get_user_chunks, show_users_allocation
from Xpress_parse.labels import get_labels
from Xpress_parse.collect import select_records
from Xpress_parse.progress import read_past_progress
from Xpress_parse.write import get_output_fp, write_selection


def xpress(**args) -> None:
    """
    Main script to collect the records, parse them and interact with the
    user to make a selection that potentially can be labelled.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)
            output_fp: str
                Selected records (default to xpress_records_YYYY-MM-DD.tsv)
            user : tuple
                Username(s)
            labels_fp : str
                Path to file with 2 tab-separated columns: keywords and labels
    """
    # Get the record tables by concatenating all input record tables
    record_pd = get_records(args)
    # Get the user work loads is a series of users is specified
    user = get_user_chunks(args, record_pd)
    # Get the labels from file or from user interactively
    labels = get_labels(args)
    # Check whether some selection was already made and if yes collect progress
    selected, progress = read_past_progress(args)
    # Print the number of records allocate to each user
    show_users_allocation(record_pd, user, selected)
    input('\nPress any key to start...')
    # Make selection interactively
    cur_time = datetime.now().strftime("%Y%m%d%H%M%S")
    select_records(record_pd, user, labels, selected, progress, cur_time)
    # write the selection
    output_fp = get_output_fp(args, user, cur_time[:8])
    write_selection(output_fp, record_pd, selected)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import unittest
from unittest.mock import patch
import pkg_resources
from os.path import isfile

from Xpress_parse.collect import *


class TestCollect(unittest.TestCase):

    def setUp(self) -> None:
        self.record_pd = pd.DataFrame(
            [['1', 'f1', 'she'], ['2', 'f2', 'they'], ['3', 'f3', 'you']],
            columns=['AB', 'file', 'PERSON'], index=[0, 1, 2])

        self.selected = {'record#2': (True, ['no_label'])}

        self.query = '>>> Exit\t: write "exit", "stop" or "quit" + press '
        self.query += '<Return>\n>>> Discard\t: press <Return>\n>>> Keep\t'
        self.query += ': write anything (or a keyword) + press <Return>'

        test_folder = pkg_resources.resource_filename('Xpress_parse', 'test')
        self.progress_dir = '%s/progress_folder' % test_folder
        self.progress = '%s/rad' % self.progress_dir
        self.progress_fp = '%s_TIME.tsv' % self.progress

        self.row = pd.Series(['a', 'b', 'c'], index=['AU', 'TI', 'AB'])

    @patch('builtins.input', return_value='')
    def test_user_interact_no_return(self, mock_input):
        obs = user_interact(self.record_pd, self.selected, self.progress,
                            {'1': 'a'}, 'TIME', 0, 0, 1, self.row)
        exp = (False, None)
        self.assertEqual(exp, obs)

    @patch('builtins.input', return_value='5')
    def test_user_interact_no_label(self, mock_input):
        obs = user_interact(self.record_pd, self.selected, self.progress,
                            {'1': 'a'}, 'TIME', 0, 0, 1, self.row)
        exp = (True, ['no_label'])
        self.assertEqual(exp, obs)

    @patch('builtins.input', return_value='1')
    def test_user_interact_label(self, mock_input):
        obs = user_interact(self.record_pd, self.selected, self.progress,
                            {'1': 'a'}, 'TIME', 0, 0, 1, self.row)
        exp = (True, ['a'])
        self.assertEqual(exp, obs)

    # the two last '' are to end the loop are the user is asked to add/remove
    @patch('builtins.input', side_effect=['exit', '', ''])
    def test_user_interact(self, mock_input):
        with self.assertRaises(SystemExit) as cm:
            user_interact(self.record_pd, self.selected, self.progress,
                          {'1': 'a'}, 'TIME', 0, 0, 1, self.row)
        self.assertEqual(cm.exception.code, 0)

    @patch('builtins.input', side_effect=['exit', '0', '1', '2', '', '1', ''])
    def test_user_interact_mixed(self, mock_input):
        with self.assertRaises(SystemExit) as cm:
            user_interact(self.record_pd, self.selected, self.progress,
                          {'1': 'a'}, 'TIME', 0, 0, 1, self.row)
        self.assertEqual(cm.exception.code, 0)
        exp = {'record#0': (True, ['no_label']),
               'record#1': (False, None),
               'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)

    @patch('builtins.print')
    def test_write_out_progress(self, mock_print):
        write_out_progress({'1': (False, None)}, self.progress, 'TIME')
        with open(self.progress_fp) as f:
            for line in f:
                obs = line.strip()
        self.assertEqual('{"1": [false, null]}', obs)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(
            print_out,
            'Use the same exact command line to restart from here. Goodbye.')

        select = {'1': (False, None), '2': (True, ['a', 'b'])}
        write_out_progress(select, self.progress, 'TIME')
        with open(self.progress_fp) as f:
            for line in f:
                obs = line.strip()
        self.assertEqual('{"1": [false, null], "2": [true, ["a", "b"]]}', obs)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(
            print_out,
            'Use the same exact command line to restart from here. Goodbye.')

    @patch('builtins.input', return_value='')
    def test_ask_it_add_no_ret(self, mock_input):
        obs = ask_it('add', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)
        self.assertEqual('', obs)
        obs = ask_it('remove', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)
        self.assertEqual('', obs)

    @patch('builtins.input', return_value='not_in')
    def test_ask_it_add_not_in(self, mock_input):
        obs = ask_it('add', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)
        self.assertEqual('not_in', obs)

    @patch('builtins.input', return_value='not_in')
    def test_ask_it_remove_not_in(self, mock_input):
        obs = ask_it('add', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)
        self.assertEqual('not_in', obs)

    @patch('builtins.input', return_value='1')
    def test_ask_it_add(self, mock_input):
        obs = ask_it('add', self.record_pd, self.selected)
        exp = {'record#2': (True, ['no_label']),
               'record#1': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)
        self.assertEqual('1', obs)

    @patch('builtins.input', return_value='1')
    def test_ask_it_change(self, mock_input):
        obs = ask_it('remove', self.record_pd, self.selected)
        exp = {'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)
        self.assertEqual('1', obs)

    @patch('builtins.input', return_value='1')
    def test_ask_it_remove_nothing(self, mock_input):
        obs = ask_it('remove', self.record_pd, self.selected)
        exp = {'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)
        self.assertEqual('1', obs)

    @patch('builtins.input', return_value='2')
    def test_ask_it_remove(self, mock_input):
        obs = ask_it('remove', self.record_pd, self.selected)
        exp = {'record#2': (False, None)}
        self.assertEqual(exp, self.selected)
        self.assertEqual('2', obs)

    @patch('builtins.input', side_effect=['not_in', ''])
    def test_adding_something_add_no_ret(self, mock_input):
        adding_something('add', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)

    @patch('builtins.input', side_effect=['not_in', ''])
    def test_adding_something_remove_no_ret(self, mock_input):
        adding_something('remove', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)

    @patch('builtins.input', side_effect=['5', ''])
    def test_adding_something_add_not_in(self, mock_input):
        adding_something('add', self.record_pd, self.selected)
        self.assertEqual({'record#2': (True, ['no_label'])}, self.selected)

    @patch('builtins.input', side_effect=['1', ''])
    def test_adding_something_add(self, mock_input):
        adding_something('add', self.record_pd, self.selected)
        exp = {'record#2': (True, ['no_label']),
               'record#1': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['1', ''])
    def test_adding_something_change(self, mock_input):
        adding_something('remove', self.record_pd, self.selected)
        exp = {'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['1', ''])
    def test_adding_something_remove_nothing(self, mock_input):
        adding_something('remove', self.record_pd, self.selected)
        exp = {'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['2', ''])
    def test_adding_something(self, mock_input):
        adding_something('remove', self.record_pd, self.selected)
        exp = {'record#2': (False, None)}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['0', '1', '3', '4', ''])
    def test_adding_something_add_multiple(self, mock_input):
        adding_something('add', self.record_pd, self.selected)
        exp = {'record#0': (True, ['no_label']),
               'record#1': (True, ['no_label']),
               'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['0', '1', '2', '3', '4', ''])
    def test_adding_something_remove_multiple(self, mock_input):
        adding_something('remove', self.record_pd, self.selected)
        exp = {'record#2': (False, None)}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['0', '1', '', '2', '3', '4', ''])
    def test_add_and_remove(self, mock_input):
        self.selected = {'record#2': (True, ['no_label'])}
        add_and_remove(self.record_pd, self.selected)
        exp = {'record#0': (True, ['no_label']),
               'record#1': (True, ['no_label']),
               'record#2': (False, None)}
        self.assertEqual(exp, self.selected)

    @patch('builtins.input', side_effect=['0', '1', '2', '', '1', ''])
    def test_add_and_remove_mixed(self, mock_input):
        self.selected = {'record#2': (True, ['no_label'])}
        add_and_remove(self.record_pd, self.selected)
        exp = {'record#0': (True, ['no_label']),
               'record#1': (False, None),
               'record#2': (True, ['no_label'])}
        self.assertEqual(exp, self.selected)

    def test_get_query(self):
        obs = get_query({'a': '1'})
        self.query += '\n    [ Your labels: a -> 1 ]\n>>> '
        self.assertEqual(self.query, obs)

    def test_get_query2(self):
        obs = get_query({'a': '1', 'b': 2})
        self.query += '\n    [ Your labels: a -> 1'
        self.query += '\n                   b -> 2 ]\n>>> '
        self.assertEqual(self.query, obs)

    def tearDown(self) -> None:
        if isfile(self.progress_fp):
            os.remove(self.progress_fp)
        if isdir(self.progress_dir):
            os.rmdir(self.progress_dir)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import json
import glob
import hashlib
from os.path import dirname, isdir


def get_progress_files(
        progress: str,
        selected: dict) -> None:
    """Fill the selected object with the latest records selection.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)
    progress : str
        Path to the file (no extension) that will contain the progress
        for the currently parsed WoS searches
    selected : dict
        Selected records
    """
    to_glob = '%s_2*.tsv' % progress
    fps = sorted(glob.glob(to_glob))
    if len(fps):
        file_handle = open(fps[-1], "r")
        selected.update(json.load(file_handle))
        file_handle.close()


def get_progress_base(args: dict) -> str:
    """Get the folder and basename where the progress data will be kept updated.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)

    Returns
    -------
    progress : str
        Path to the file (no extension) that will contain the progress
        for the currently parsed WoS searches
    """
    input_fps = args['input_fps']
    folder = '%s/.xpress' % dirname(input_fps[0])
    base = hashlib.sha224(''.join(input_fps).encode()).hexdigest()
    progress = folder + '/' + base
    return progress


def read_past_progress(args: dict) -> tuple:
    """Initialization checks whether the output file already exists which
    means you already made selection and stopped.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)

    Returns
    -------
    selected : dict
        Selected records
    progress : str
        Path to the file (no extension) that will contain the progress
        for the currently parsed WoS searches
    """
    progress = get_progress_base(args)
    selected = {}
    if isdir(dirname(progress)):
        get_progress_files(progress, selected)
    return selected, progress



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd
from os.path import isfile


def curate_labels(labels: dict) -> None:
    """Removed keyword from the labels that are reserved to leave the parsing.

    Parameters
    ----------
    labels : dict
        Mapping from keywords (keys) to labels (values)
    """
    if 'exit' in labels:
        print('Removing "exit" form labels (reserved to exit...)')
        del labels['exit']
    if 'quit' in labels:
        print('Removing "quit" form labels (reserved to exit...)')
        del labels['quit']
    if 'stop' in labels:
        print('Removing "quit" form labels (reserved to exit...)')
        del labels['stop']


def get_labels(args: dict) -> dict:
    """Get the labels from file or from user interactively.

    Parameters
    ----------
    args : dict
        All arguments, including:
            labels_fp : str
                Path to file with 2 tab-separated columns: keywords and labels

    Returns
    -------
    labels : dict
        Mapping from keywords (keys) to labels (values)
    """
    labels_fp = args['labels_fp']
    if labels_fp and isfile(labels_fp):
        labels = read_labels(labels_fp)
    else:
        labels = collect_labels()
    curate_labels(labels)
    return labels


def read_labels(labels_fp: str) -> dict:
    """Read labels from file.

    Parameters
    ----------
    labels_fp : str
        Path to file with 2 tab-separated columns: keywords and labels

    Returns
    -------
    labels : dict
        Mapping from keywords (keys) to labels (values)
    """
    labels_pd = pd.read_csv(labels_fp, header=None, sep='\t', usecols=[0, 1])
    labels = dict(labels_pd.values)
    return labels


def collect_labels() -> dict:
    """Get the labels from user interactively.

    Returns
    -------
    labels : dict
        Mapping from keywords (keys) to labels (values)
    """
    ret, paired, v = '', 0, []
    print('\n%s\nDefine keywords to label studies?\n%s' % ('-'*33, '-'*33))
    while 1:
        if paired % 2 == 0:
            key_label = '[<ENTER> to stop] keyword: '
        else:
            key_label = '> label for "%s": ' % v[-1]
        ret = input(key_label)
        if not ret:
            break
        v.append(ret)
        paired += 1
    labels = dict((y, v[x + 1]) for x, y in enumerate(v[:-1]) if x % 2 == 0)
    return labels



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import sys
import json
import pandas as pd
from os.path import dirname, isdir


def ask_it(
        remove_add: str,
        record_pd: pd.DataFrame,
        selected: dict) -> str:
    """Ask the user whether a record is to be kept
    to (or removed from) the selection or not.

    Parameters
    ----------
    remove_add : str
        Whether to remove or add the record
    record_pd : pd.DataFrame
        Records table
    selected : dict
        Whether each record (key) is relevant (value = 1) or not (value = 0)

    Returns
    -------
    ret : str
        The string entered by the user for the current abstract
    """
    ret = input('Any record to %s? If yes, enter record number (e.g. 39 to %s '
                '"record#39") and press <Return>; if not, press <Return>\n' % (
                    remove_add, remove_add))
    if ret:
        record = 'record#%s' % ret
        if str(ret) in set(record_pd.index.astype(str)):
            if remove_add == 'add':
                selected[record] = (True, ['no_label'])
            elif record in selected:
                selected[record] = (False, None)
        else:
            print('\nRecord number "%s" not valid' % ret)

    return ret


def adding_something(
        remove_add: str,
        record_pd: pd.DataFrame,
        selected: dict) -> None:
    """Keeps asking the user whether the record is to be kept or removed.

    Parameters
    ----------
    remove_add : str
        Whether to remove or add the record
    record_pd : pd.DataFrame
        Records table
    selected : dict
        Whether each record (key) is relevant (value = 1) or not (value = 0)
    """
    ret = 'x'
    while ret:
        ret = ask_it(remove_add, record_pd, selected)


def write_out_progress(
        selected: dict,
        progress: str,
        cur_time: str) -> None:
    """Write the current progress in the file associated to the current search.

    Parameters
    ----------
    selected : dict
        Selected records
    progress : str
        Path to the file (no extension) that will contain the progress
        for the currently parsed WoS searches
    cur_time : str
        Formatted data and time of now
    """
    o_fp = '%s_%s.tsv' % (progress, cur_time)
    if not isdir(dirname(progress)):
        os.makedirs(dirname(progress))
    with open(o_fp, 'w') as o_json:
        json.dump(selected, o_json)
    print('Use the same exact command line to restart from here. Goodbye.')


def fix_version(selected: dict):
    """Update

    Parameters
    ----------
    selected : dict
        Selected studies

    Returns
    -------
    selected : dict
        Selected studies updated
    """
    selected = dict(
        (x, y) if isinstance(y, list) else (x, [y, 'not_available'])
        for x, y in selected.items())
    return selected


def make_progress_dir(progress_dir: str) -> bool:
    """Get and potentially create a directory that will
    contain the progress content.

    Parameters
    ----------
    progress_dir : str
        Path to the folder that will contain the progress for the
        currently parsed WoS searches

    Returns
    -------
    bool
        Whether the progress folder exists for the current search
    """
    if not isdir(progress_dir):
        os.makedirs(progress_dir)
        return False
    else:
        return True


def clear_output():
    """Clear the stdin and stdout on the Windows or Unix terminal"""
    os.system('cls' if os.name == 'nt' else 'clear')


def get_query(labels: dict) -> str:
    """Get the query to be shown to the user under a record abstract.

    Parameters
    ----------
    labels : dict
        Mapping from keywords (keys) to labels (values)

    Returns
    -------
    query : str
        The query shown to the user
    """
    query = '>>> Exit\t: write "exit", "stop" or "quit" + press <Return>\n'
    query += '>>> Discard\t: press <Return>\n'
    query += '>>> Keep\t: write anything (or a keyword) + press <Return>'
    if labels:
        your_labels = '    [ Your labels:'
        for idx, (k, v) in enumerate(sorted(labels.items())):
            if idx:
                query += '\n%s %s -> %s' % (' ' * len(your_labels), k, v)
            else:
                query += '\n%s %s -> %s' % (your_labels, k, v)
        query += ' ]'
    query += '\n>>> '
    return query


def user_interact(
        user_record_pd: pd.DataFrame,
        selected: dict,
        progress: str,
        labels: dict,
        cur_time: str,
        rdx: int,
        r: int,
        n: int,
        row: pd.Series) -> tuple:
    """Collect the user input for a record.

    Parameters
    ----------
    user_record_pd : pd.DataFrame
        Table of the WoS search result(s) reduced to the records
        assigned to the current user
    selected : dict
        Selected records
    progress : str
        Path to the file (no extension) that will contain the progress
        for the currently parsed WoS searches
    labels : dict
        Mapping from keywords (keys) to labels (values)
    cur_time : str
        Formatted data and time of now
    rdx : int
        Iteration index
    r : int
        Index of the record in the original concatenated table
    n : int
        Total number of records for the current username
    row : pd.Series
        Record data

    Returns
    -------
    ret_val : tuple
        (bool, list) for whether to keep the record and with what labels
    """
    query = get_query(labels)
    ret = input('record#%s (%s/%s):\n%s\n%s et al. %s\n%s\n%s\n\n%s' % (
        r, rdx, n,
        '=' * 21,
        row['AU'].split(',')[0], row['TI'],
        '-' * 5,
        row['AB'],
        query
    ))

    if ret.lower() in ['exit', 'stop', 'quit']:
        # stopping asks to keep/remove records and writes progress
        add_and_remove(user_record_pd, selected)
        write_out_progress(selected, progress, cur_time)
        sys.exit(0)

    ret_val = (False, None)
    if ret:
        ret_val = (True, [labels.get(x, 'no_label') for x in str(ret).split()])
    return ret_val


def add_and_remove(user_record_pd, selected):
    """Ask the user whether the record is to be kept or removed

    Parameters
    ----------
    user_record_pd : pd.DataFrame
        Table of the WoS search result(s) reduced to the records
        assigned to the current user
    selected : dict
        Selected records
    """
    for add_remove in ['add', 'remove']:
        adding_something(add_remove, user_record_pd, selected)


def select_records(
        record_pd: pd.DataFrame,
        user: str,
        labels: dict,
        selected: dict,
        progress: str,
        cur_time: str) -> dict:
    """Interactively parsed the abstract and for each one, ask the user
    whether to stop, discard or keep it.
        If "stopped", then the progress so
    far is kept by writing a file (that is specific) for this search.
        If "discard", the abstract is labeled as False.
        If "kept", the abstract is labeled as True and potentially obtains
        one or more user-defined labels.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Table of the WoS search result(s). A column indicating the source
        file name is added
    user : str
        Current username
    labels : dict
        Mapping from keywords (keys) to labels (values)
    selected : dict
        Selected records
    progress : str
        Path to the file (no extension) that will contain the progress
        for the currently parsed WoS searches
    cur_time : str
        Formatted data and time of now
    """
    # Subset the table to the records assigned to the current user
    user_record_pd = record_pd.loc[record_pd['PERSON'] == user]
    n = user_record_pd.shape[0]
    # Start parsing through the records
    for rdx, (r, row) in enumerate(user_record_pd.iterrows()):
        try:
            if 'record#%s' % r in selected:
                continue
            clear_output()
            selected['record#%s' % r] = user_interact(
                user_record_pd,
                selected,
                progress,
                labels,
                cur_time,
                rdx,
                r,
                n,
                row)
            clear_output()
        # using ctrl-c to exit asks to keep/remove records and writes progress
        except KeyboardInterrupt:
            add_and_remove(user_record_pd, selected)
            write_out_progress(selected, progress, cur_time)
            sys.exit(0)
    return selected



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import unittest
import pkg_resources

from Xpress_parse.progress import *


class Test(unittest.TestCase):

    def setUp(self) -> None:
        test_folder = pkg_resources.resource_filename('Xpress_parse', 'test')
        self.selected = {}
        self.prog = '%s/rad' % test_folder
        self.fp1 = '%s_21.tsv' % self.prog
        self.fp2 = '%s_22.tsv' % self.prog
        self.input_fps = (self.fp1, self.fp2,)
        with open(self.fp1, 'w') as o:
            o.write('{"record#0": [false, null], "record#1": [false, "a"]}')
        with open(self.fp2, 'w') as o:
            o.write('{"record#2": [false, null], "record#1": [false, "b"]}')

        self.args = {'input_fps': ('/a/b.tsv', '/c/d.tsv',)}
        self.hashlib = hashlib.sha224('/a/b.tsv/c/d.tsv'.encode()).hexdigest()
        self.prog_dir = '/a/.xpress/' + self.hashlib

    def test_get_progress_files(self):
        get_progress_files(self.prog, self.selected)
        exp = {"record#2": [False, None], "record#1": [False, "b"]}
        self.assertEqual(self.selected, exp)

    def test_get_progress_files_fp1(self):
        os.remove(self.fp2)  # suppress most recent progress
        get_progress_files(self.prog, self.selected)
        exp = {"record#0": [False, None], "record#1": [False, "a"]}
        self.assertEqual(self.selected, exp)

    def test_get_progress_files_none(self):
        os.remove(self.fp2)  # suppress all
        os.remove(self.fp1)  # suppress all
        get_progress_files(self.prog, self.selected)
        exp = {}
        self.assertEqual(exp, self.selected)

    def test_get_progress_base(self):
        obs = get_progress_base(self.args)
        self.assertEqual(self.prog_dir, obs)

    def test_read_past_progress(self):
        selected, progress = read_past_progress(self.args)
        self.assertEqual(self.prog_dir, progress)
        self.assertEqual({}, selected)

    def test_read_past_progress_sel(self):
        selected, progress = read_past_progress(self.args)
        self.assertEqual(self.prog_dir, progress)

    def tearDown(self) -> None:
        if os.path.isfile(self.fp1):
            os.remove(self.fp1)
        if os.path.isfile(self.fp2):
            os.remove(self.fp2)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import unittest
from unittest.mock import patch
import pkg_resources

from Xpress_parse.labels import *


class TestLabels(unittest.TestCase):

    def setUp(self) -> None:
        self.root = pkg_resources.resource_filename('Xpress_parse', 'test')
        self.tsv = '%s/labels.tsv' % self.root
        self.labels = {1: 'a', 2: 'b', 3: 'c'}
        with open(self.tsv, 'w') as o:
            for k, v in self.labels.items():
                o.write('%s\t%s\n' % (k, v))
        self.tsv_extra = '%s/labels_extra.tsv' % self.root
        with open(self.tsv_extra, 'w') as o:
            for k, v in self.labels.items():
                o.write('%s\t%s\tuseless\n' % (k, v))

    def test_read_labels(self):
        obs = read_labels(self.tsv)
        self.assertEqual(self.labels, obs)
        obs = read_labels(self.tsv_extra)
        self.assertEqual(self.labels, obs)

    @patch('builtins.input', side_effect=['1', '2', '3', '4', ''])
    def test_collect_labels_1(self, input):
        obs = collect_labels()
        self.assertEqual({'1': '2', '3': '4'}, obs)

    @patch('builtins.input', side_effect=['1', '2', '3', ''])
    def test_collect_labels_2(self, input):
        obs = collect_labels()
        self.assertEqual({'1': '2'}, obs)

    @patch('builtins.input', side_effect=[''])
    def test_collect_labels_3(self, input):
        obs = collect_labels()
        self.assertEqual({}, obs)

    @patch('builtins.input', side_effect=['1', '2', '1', '3', ''])
    def test_collect_labels_4(self, input):
        obs = collect_labels()
        self.assertEqual({'1': '3'}, obs)

    def test_curate_labels(self):
        labels = {'what': 'ever'}
        exp = {'what': 'ever'}
        curate_labels(labels)
        self.assertEqual(labels, exp)

        labels = {}
        exp = {}
        curate_labels(labels)
        self.assertEqual(labels, exp)

        labels = {'exit': 'is_wrong'}
        exp = {}
        curate_labels(labels)
        self.assertEqual(labels, exp)

        labels = {'exit': 'is_wrong', 'quit': 'is_wrong'}
        exp = {}
        curate_labels(labels)
        self.assertEqual(labels, exp)

        labels = {'exit': 'is_wrong', 'quit': 'is_wrong', 'stop': 'is_wrong'}
        exp = {}
        curate_labels(labels)
        self.assertEqual(labels, exp)

    def tearDown(self) -> None:
        os.remove(self.tsv)
        os.remove(self.tsv_extra)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd
from os.path import isfile


def read_records_fp(input_fp: str) -> pd.DataFrame:
    """
    Read one file of records returned by a WoS query.

    Parameters
    ----------
    input_fp : str
        Path to one WoS records file

    Returns
    -------
    records_pd : pd.DataFrame
        Table of the records
    """
    records_read = open(input_fp)
    records_lines = records_read.readlines()
    records_read.close()
    records = [x.strip().split('\t') for x in records_lines]
    records = pd.DataFrame(records[1:], columns=records[0])
    return records


def read_records_fps(args: dict) -> pd.DataFrame:
    """Read the WoS search results as tables and concatenate
    these into one main table.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)

    Returns
    -------
    record_pd : pd.DataFrame
        Table of the WoS search result(s)
        A column indicating the source file name is added
    """
    records = []
    for input_fp in args['input_fps']:
        if not isfile(input_fp):
            raise IOError('%s does not exist' % input_fp)
        record_pd = read_records_fp(input_fp)
        record_pd['file'] = input_fp
        records.append(record_pd)
    record_pd = pd.concat(records)
    record_pd.index = range(len(record_pd))
    return record_pd


def show_file_per_record(record_pd: pd.DataFrame) -> None:
    """Show the number of files in which the records may be shared.
    This is information before dropping these duplicates. Note that the file
    origin information will remain tracked.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Table of the WoS search result(s)
        A column indicating the source file name is added
    """
    count_pd = record_pd[['AB', 'file']].groupby('AB').count()
    n = record_pd['file'].nunique()
    vc = count_pd.value_counts()
    if int(count_pd.max()) > 1:
        print('\n%s\nRedundant records found:\n%s' % ('-'*24, '-'*24))
        for i, j in dict(vc).items():
            m = '- %s records present in %s file' % (j, i[0])
            if i[0] > 1:
                m += 's'
            print(m)
    elif n > 1:
        print('\n%s\nRecords unique to the %s files\n%s' % ('-'*30, n, '-'*30))


def remove_empty_abstract(record_pd: pd.DataFrame) -> None:
    """Remove the row of the table that do not have an abstract and
    thus that will be skipped. This also reset the index.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Table of the WoS search result(s)
    """
    to_drop = [x for x, row in record_pd.iterrows() if row['AB'] == '']
    if to_drop:
        record_pd.drop(index=to_drop, inplace=True)


def show_no_abstracts(record_pd: pd.DataFrame) -> None:
    """Show the recirds that do not have an abstract and thus that will be
    skipped.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Table of the WoS search result(s)
        A column indicating the source file name is added

    Returns
    -------
    record_pd : pd.DataFrame
        Same table reduced to those records with available abstract
    """
    empty_pd = record_pd.loc[record_pd['AB'] == ''].copy()
    if empty_pd.shape[0]:
        print('\n%s\n%s abstracts empty:\n%s' % ('-'*19, len(empty_pd), '-'*19))
        no_record_dp = empty_pd.sort_values('PY')
        for rdx, (_, row) in enumerate(no_record_dp.iterrows()):
            print('[%s] %s, %s: %s' % (rdx, row['AU'].split(',')[0],
                                       row['PY'], row['TI'].lower()))


def update_file(record_pd: pd.DataFrame) -> None:
    """Updates the content of the "file" column so that abstracts
    present in multiple files have the full set of files.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Same table reduced to those records with available abstract
    """
    file_per_ab = record_pd[['AB', 'file']].groupby('AB').apply(
        lambda x: ','.join(sorted(set(x['file'].values)))).to_dict()
    record_pd['file'] = [file_per_ab[x] for x in record_pd['AB']]
    record_pd.drop_duplicates(inplace=True)
    record_pd.index = range(len(record_pd))


def get_records(args: dict) -> pd.DataFrame:
    """Get the record tables by concatenating all input record tables.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)
            output_fp: str
                Selected records (default to xpress_records_YYYY-MM-DD.tsv)
            user : tuple
                User name(s)

    Returns
    -------
    record_pd : pd.DataFrame
        Table of the WoS search result(s). A column indicating the source
        file name is added
    """
    record_pd = read_records_fps(args)
    show_no_abstracts(record_pd)
    remove_empty_abstract(record_pd)
    show_file_per_record(record_pd)
    update_file(record_pd)
    return record_pd



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import unittest
from unittest.mock import patch

from pandas.testing import assert_frame_equal
from Xpress_parse.users import *


class TestUsers(unittest.TestCase):

    def setUp(self) -> None:
        self.args = {'user': ['you', 'they', 'she']}
        self.record_pd = pd.DataFrame(
            [['1', 'file1'], ['2', 'file2'], ['3', 'file3']],
            columns=['AB', 'file'])
        self.record_user_pd = pd.DataFrame(
            [['1', 'file1', 'she'], ['2', 'file2', 'they'],
             ['3', 'file3', 'you']],
            columns=['AB', 'file', 'PERSON'])
        self.record_no_user_pd = pd.DataFrame(
            [['1', 'file1', ''], ['2', 'file2', ''], ['3', 'file3', '']],
            columns=['AB', 'file', 'PERSON'])

        self.args2 = {'user': ['she', 'you', 'they']}
        self.record2_pd = pd.DataFrame(
            [['1', 'file1'], ['2', 'file2'], ['3', 'file3'],
             ['4', 'file4'], ['5', 'file5'], ['6', 'file6'],
             ['7', 'file7'], ['8', 'file8']],
            columns=['AB', 'file'])
        self.record2_user_pd = pd.DataFrame(
            [['1', 'file1', 'she'], ['2', 'file2', 'she'],
             ['3', 'file3', 'she'], ['4', 'file4', 'they'],
             ['5', 'file5', 'they'], ['6', 'file6', 'you'],
             ['7', 'file7', 'you'], ['8', 'file8', 'you']],
            columns=['AB', 'file', 'PERSON'])
        self.record2_no_user_pd = pd.DataFrame(
            [['1', 'file1', ''], ['2', 'file2', ''], ['3', 'file3', ''],
             ['4', 'file4', ''], ['5', 'file5', ''], ['6', 'file6', ''],
             ['7', 'file7', ''], ['8', 'file8', '']],
            columns=['AB', 'file', 'PERSON'])

        self.args3 = {'user': ['they', 'you', 'she', 'she', 'she']}
        self.record3_pd = pd.DataFrame(
            [['1', 'file1'], ['2', 'file2'], ['3', 'file3'], ['4', 'file4']],
            columns=['AB', 'file'])
        self.record3_user_pd = pd.DataFrame(
            [['1', 'file1', 'she'], ['2', 'file2', 'she'],
             ['3', 'file3', 'they'], ['4', 'file4', 'you']],
            columns=['AB', 'file', 'PERSON'])
        self.record3_no_user_pd = pd.DataFrame(
            [['1', 'file1', ''], ['2', 'file2', ''], ['3', 'file3', ''],
             ['4', 'file4', '']],
            columns=['AB', 'file', 'PERSON'])

        hyphens = ('-' * 29, '-' * 29)
        self.exp = "\n%s\nRecords allocated per person:\n%s" % hyphens

    def test_add_person_column(self):
        add_person_column(self.args, self.record_pd)
        assert_frame_equal(self.record_pd, self.record_user_pd)
        add_person_column(self.args, self.record2_pd)
        assert_frame_equal(self.record2_pd, self.record2_user_pd)
        add_person_column(self.args, self.record3_pd)
        assert_frame_equal(self.record3_pd, self.record3_user_pd)

    def test_get_user(self):
        obs = get_user(self.args, self.record_pd)
        self.assertEqual(obs, 'you')
        obs = get_user(self.args2, self.record2_pd)
        self.assertEqual(obs, 'she')
        obs = get_user(self.args3, self.record3_pd)
        self.assertEqual(obs, 'they')

    def test_no_user(self):
        obs = no_user(self.record_pd)
        assert_frame_equal(self.record_pd, self.record_no_user_pd)
        self.assertEqual('', obs)
        obs = no_user(self.record2_pd)
        assert_frame_equal(self.record2_pd, self.record2_no_user_pd)
        self.assertEqual('', obs)
        obs = no_user(self.record3_pd)
        assert_frame_equal(self.record3_pd, self.record3_no_user_pd)
        self.assertEqual('', obs)


class TestUSerAlloc(unittest.TestCase):

    def setUp(self) -> None:
        self.record_pd = pd.DataFrame(
            [['1', 'file1', 'she'],
             ['2', 'file2', 'they'],
             ['3', 'file3', 'you']],
            columns=['AB', 'file', 'PERSON'])
        self.record2_pd = pd.DataFrame(
            [['1', 'file1', 'she'],
             ['2', 'file2', 'she'],
             ['3', 'file3', 'she'],
             ['4', 'file4', 'they'],
             ['5', 'file5', 'they'],
             ['6', 'file6', 'you'],
             ['7', 'file7', 'you'],
             ['8', 'file8', 'you']],
            columns=['AB', 'file', 'PERSON'])
        self.record3_pd = pd.DataFrame(
            [['1', 'file1', 'she'],
             ['2', 'file2', 'she'],
             ['3', 'file3', 'they'],
             ['4', 'file4', 'you']],
            columns=['AB', 'file', 'PERSON'])
        self.selected = {'1': (True, ['whatever'])}
        hyphens = ('-' * 29, '-' * 29)
        self.exp = "\n%s\nRecords allocated per person:\n%s" % hyphens

    @patch('builtins.print')
    def test_show_users_allocation(self, mock_print):
        show_users_allocation(self.record_pd, 'you', self.selected)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.exp)
        print_out = mock_print.call_args_list[1][0][0]
        self.assertEqual(print_out, '- she: 1 records')
        print_out = mock_print.call_args_list[2][0][0]
        self.assertEqual(print_out, '- they: 1 records')
        print_out = mock_print.call_args_list[3][0][0]
        self.assertEqual(print_out,
                         '- you: 1 records (you: 1 records already parsed)')

    @patch('builtins.print')
    def test_show_users_allocation2(self, mock_print):
        show_users_allocation(self.record2_pd, 'you', self.selected)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.exp)
        print_out = mock_print.call_args_list[1][0][0]
        self.assertEqual(print_out, '- she: 3 records')
        print_out = mock_print.call_args_list[2][0][0]
        self.assertEqual(print_out, '- they: 2 records')
        print_out = mock_print.call_args_list[3][0][0]
        self.assertEqual(print_out,
                         '- you: 3 records (you: 1 records already parsed)')

    @patch('builtins.print')
    def test_show_users_allocation3(self, mock_print):
        show_users_allocation(self.record3_pd, 'she', self.selected)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.exp)
        print_out = mock_print.call_args_list[1][0][0]
        self.assertEqual(print_out,
                         '- she: 2 records (you: 1 records already parsed)')
        print_out = mock_print.call_args_list[2][0][0]
        self.assertEqual(print_out, '- they: 1 records')
        print_out = mock_print.call_args_list[3][0][0]
        self.assertEqual(print_out, '- you: 1 records')

    @patch('builtins.print')
    def test_show_no_users_allocation(self, mock_print):
        show_users_allocation(self.record_pd, '', self.selected)
        print_out = mock_print.call_args_list[0][0][0]
        exp = "\n%s\nNumber of records to parse: 3\n%s" % ('-' * 30, '-' * 30)
        self.assertEqual(print_out, exp)

    @patch('builtins.print')
    def test_show_no_users_allocation2(self, mock_print):
        show_users_allocation(self.record2_pd, '', self.selected)
        print_out = mock_print.call_args_list[0][0][0]
        exp = "\n%s\nNumber of records to parse: 8\n%s" % ('-' * 30, '-' * 30)
        self.assertEqual(print_out, exp)

    @patch('builtins.print')
    def test_show_no_users_allocation3(self, mock_print):
        show_users_allocation(self.record3_pd, '', self.selected)
        print_out = mock_print.call_args_list[0][0][0]
        exp = "\n%s\nNumber of records to parse: 4\n%s" % ('-' * 30, '-' * 30)
        self.assertEqual(print_out, exp)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import unittest
from unittest.mock import patch

import os
import pkg_resources
from pandas.testing import assert_frame_equal

from Xpress_parse.records import *


class TestUpdateFile(unittest.TestCase):

    def setUp(self) -> None:
        self.pd_redundancy = pd.DataFrame(
            [['1', 'file1a'],       # same 4 abstracts "1"
                ['1', 'file2'],
                ['1', 'file3'],
                ['1', 'file1b'],
             ['2', 'file2'],        # one abstract "2"
             ['3', 'file3a'],       # same 3 abstracts "3"
                ['3', 'file3b'],
                ['3', 'file3c']],
            columns=['AB', 'file'])
        # no redundancy
        self.pd_no_redundancy = pd.DataFrame(
            [['1', 'file1'],
             ['2', 'file2'],
             ['3', 'file3']],
            columns=['AB', 'file'])

    def test_update_file_redundancy(self):
        update_file(self.pd_redundancy)
        exp = pd.DataFrame(
            [['1', 'file1a,file1b,file2,file3'],
             ['2', 'file2'],
             ['3', 'file3a,file3b,file3c']],
            columns=['AB', 'file'])
        assert_frame_equal(self.pd_redundancy, exp)

    def test_update_file_no_redundancy(self):
        update_file(self.pd_no_redundancy)
        exp = pd.DataFrame(
            [['1', 'file1'],
             ['2', 'file2'],
             ['3', 'file3']],
            columns=['AB', 'file'])
        assert_frame_equal(self.pd_no_redundancy, exp)


class TestShowFilePerRecord(unittest.TestCase):

    def setUp(self) -> None:
        self.pd_single = pd.DataFrame(
            [['1', 'file1'], ['2', 'file1'], ['3', 'file1']],
            columns=['AB', 'file'])

        dashes = ('-' * 24, '-' * 24)
        self.pd_1_exp = '\n%s\nRedundant records found:\n%s' % dashes
        self.pd_1_exp_1 = '- 3 records present in 2 files'
        self.pd_1 = pd.DataFrame(
            [['1', 'file1'], ['2', 'file1'], ['3', 'file1'],
             ['1', 'file2'], ['2', 'file2'], ['3', 'file2']],
            columns=['AB', 'file'])

        dashes = ('-' * 24, '-' * 24)
        self.pd_2_exp = '\n%s\nRedundant records found:\n%s' % dashes
        self.pd_2_exp_1 = '- 3 records present in 1 file'
        self.pd_2_exp_2 = '- 2 records present in 2 files'
        self.pd_2 = pd.DataFrame(
            [['1', 'file1'], ['2', 'file1'], ['3', 'file1'], ['4', 'file1'],
             ['4', 'file2'], ['5', 'file1'], ['5', 'file3']],
            columns=['AB', 'file'])

        dashes = ('-' * 30, '-' * 30)
        self.pd_exp = '\n%s\nRecords unique to the 3 files\n%s' % dashes
        self.pd = pd.DataFrame(
            [['1', 'file1'], ['2', 'file2'], ['3', 'file3']],
            columns=['AB', 'file'])

    def test_show_file_per_record_single(self):
        self.assertIsNone(show_file_per_record(self.pd_single))

    @patch('builtins.print')
    def test_show_file_per_record_1(self, mock_print):
        show_file_per_record(self.pd_1)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.pd_1_exp)
        print_out = mock_print.call_args_list[1][0][0]
        self.assertEqual(print_out, self.pd_1_exp_1)
        with self.assertRaises(IndexError):
            print_out = mock_print.call_args_list[2][0][0]

    @patch('builtins.print')
    def test_show_file_per_record_2(self, mock_print):
        show_file_per_record(self.pd_2)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.pd_2_exp)
        print_out = mock_print.call_args_list[1][0][0]
        self.assertEqual(print_out, self.pd_2_exp_1)
        print_out = mock_print.call_args_list[2][0][0]
        self.assertEqual(print_out, self.pd_2_exp_2)
        with self.assertRaises(IndexError):
            print_out = mock_print.call_args_list[3][0][0]

    @patch('builtins.print')
    def test_show_file_per_record(self, mock_print):
        show_file_per_record(self.pd)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.pd_exp)


class TestRemoveEmptyAbstract(unittest.TestCase):

    def setUp(self) -> None:
        self.nothing = pd.DataFrame(
            [['1', 'file1'], ['2', 'file1'], ['3', 'file1']],
            columns=['AB', 'file'])
        self.after = pd.DataFrame(
            [['1', 'file1'], ['2', 'file1'], ['3', 'file1']],
            columns=['AB', 'file'])

        self.empty = pd.DataFrame(
            [['', 'file1'], ['', 'file1'], ['3', 'file1']],
            columns=['AB', 'file'])
        self.after_empty = pd.DataFrame(
            [['3', 'file1']], columns=['AB', 'file'], index=[2])

    def test_remove_empty_abstract(self):
        remove_empty_abstract(self.nothing)
        assert_frame_equal(self.nothing, self.after)
        remove_empty_abstract(self.empty)
        assert_frame_equal(self.empty, self.after_empty)


class TestShowNoAbstracts(unittest.TestCase):

    def setUp(self) -> None:
        self.missing = pd.DataFrame(
            [['', 'file1', 2, 'B,b', 'Title2'],
             ['', 'file1', 1, 'A,a', 'TITLE1'],
             ['3', 'file1', 3, 'C,c', 'Title3']],
            columns=['AB', 'file', 'PY', 'AU', 'TI'], index=[4, 1, 1])

        dashes = ('-' * 19, '-' * 19)
        self.missing_exp = '\n%s\n2 abstracts empty:\n%s' % dashes
        self.missing_exp_1 = '[0] A, 1: title1'
        self.missing_exp_2 = '[1] B, 2: title2'

    @patch('builtins.print')
    def test_show_no_abstracts(self, mock_print):
        show_no_abstracts(self.missing)
        print_out = mock_print.call_args_list[0][0][0]
        self.assertEqual(print_out, self.missing_exp)
        print_out = mock_print.call_args_list[1][0][0]
        self.assertEqual(print_out, self.missing_exp_1)
        print_out = mock_print.call_args_list[2][0][0]
        self.assertEqual(print_out, self.missing_exp_2)
        with self.assertRaises(IndexError):
            print_out = mock_print.call_args_list[3][0][0]


class TestReadRecordsFp(unittest.TestCase):

    def setUp(self) -> None:
        self.data = [['A', 'B', 'C', 'D'],
                     ['A1', 'B1', 'C1', 'D1'],
                     ['A2', 'B2', 'C2', 'D2']]
        self.root = pkg_resources.resource_filename('Xpress_parse', 'test')
        self.test_fp1 = '%s/test_fp1' % self.root
        self.test_fp2 = '%s/test_fp2' % self.root
        self.data_fps = [self.test_fp1, self.test_fp2]
        self.args = {'input_fps': self.data_fps}
        for data_fp in self.data_fps:
            with open(data_fp, 'w') as o:
                for data in self.data:
                    o.write('%s\n' % '\t'.join(data))
        self.not_exists = ['not_a_file.tsv']
        self.args_not_exists = {'input_fps': self.not_exists}

    def test_read_records_fp(self):
        obs = read_records_fp(self.data_fps[0])
        exp = pd.DataFrame([['A1', 'B1', 'C1', 'D1'],
                            ['A2', 'B2', 'C2', 'D2']],
                           columns=['A', 'B', 'C', 'D'])
        assert_frame_equal(exp, obs)

    def test_read_records_fps(self):
        obs = read_records_fps(self.args)
        exp = pd.DataFrame([['A1', 'B1', 'C1', 'D1', self.test_fp1],
                            ['A2', 'B2', 'C2', 'D2', self.test_fp1],
                            ['A1', 'B1', 'C1', 'D1', self.test_fp2],
                            ['A2', 'B2', 'C2', 'D2', self.test_fp2]],
                           columns=['A', 'B', 'C', 'D', 'file'])
        assert_frame_equal(exp, obs)

        with self.assertRaises(IOError):
            read_records_fps(self.args_not_exists)

    def tearDown(self) -> None:
        for data_fp in self.data_fps:
            os.remove(data_fp)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import numpy as np
import pandas as pd


def no_user(record_pd: pd.DataFrame) -> str:
    """Get an empty username if none was given in command line and also fill
    the "PERSON" column with empty string.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Records table

    Returns
    -------
    user : str
        Current username
    """
    record_pd['PERSON'] = ''
    user = ''
    return user


def add_person_column(args: dict, record_pd: pd.DataFrame) -> None:
    """Add the "PERSON" column to the records table that contains on name
    per record. The names being split across the listed users.

    Parameters
    ----------
    args : dict
        All arguments, including:
            user : tuple
                User name(s)
    record_pd : pd.DataFrame
        Records table
    """
    users = sorted(set(args['user']))
    # Split the list of indices corresponding to the papers
    bins = list(np.linspace(0, 1, len(users), endpoint=False))
    bins.append(1)
    # use quantile cut to split index vector in equal-sized parts
    record_pd['PERSON'] = list(pd.qcut(record_pd.index, bins, labels=users))


def get_user(args: dict, record_pd: pd.DataFrame) -> str:
    """

    Parameters
    ----------
    args : dict
        All arguments, including:
            user : tuple
                User name(s)
    record_pd : pd.DataFrame
        Records table

    Returns
    -------
    user : str
        Current user name
    """
    add_person_column(args, record_pd)
    user = args['user'][0]
    return user


def get_user_chunks(args: dict, record_pd: pd.DataFrame) -> str:
    """Get the user work loads is a series of users is specified.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)
            output_fp: str
                Selected records (default to xpress_records_YYYY-MM-DD.tsv)
            user : tuple
                User name(s)
    record_pd : pd.DataFrame
        Records table

    Returns
    -------
    user : str
        Current username
    """
    # Define the list of people that will split the load amongst themselves
    if args['user']:
        user = get_user(args, record_pd)
    else:
        user = no_user(record_pd)
    return user


def show_users_allocation(
        record_pd: pd.DataFrame,
        user: str,
        selected: dict) -> None:
    """Print the number of records allocate to each user.

    Parameters
    ----------
    record_pd : pd.DataFrame
        Records table
    user : str
        Current username
    selected : dict
        Selected studies
    """
    if not user:
        n = len(record_pd)
        print("\n%s\nNumber of records to parse: %s\n%s" % ('-'*30, n, '-'*30))
    else:
        print("\n%s\nRecords allocated per person:\n%s" % ('-' * 29, '-' * 29))
        for u, n in record_pd['PERSON'].value_counts().sort_index().iteritems():
            m = '- %s: %s records' % (u, n)
            if u == user:
                m += ' (you'
                if selected:
                    m += ': %s records already parsed' % len(selected)
                m += ')'
            print(m)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
import pandas as pd
from os.path import abspath, splitext


def get_output_fp(args: dict, user: str, cur_time: str) -> str:
    """Write the resulting selection to the output file.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fps : tuple
                Path(s) to WoS search result(s)
            output_fp: str
                Selected records (default to xpress_records_YYYY-MM-DD.tsv)
    user : str
        Current username
    cur_time : str
        Formatted data and time of now

    Returns
    -------
    output_fp : str
        Output path
    """
    if args['output_fp']:
        output_fp = args['output_fp']
    else:
        cur_date = cur_time[:8]
        output_fp = splitext(args['input_fps'][0])[0] + '_xpressed_' + cur_date
        if user:
            output_fp += '_' + user + '.tsv'
        else:
            output_fp += '.tsv'
        output_fp = abspath(output_fp)
    return output_fp


def write_selection(
        output_fp: str,
        record_pd: pd.DataFrame,
        selected: dict) -> None:
    """Write the resulting selection to the output file.

    Parameters
    ----------
    output_fp : str
        Output path
    record_pd : pd.DataFrame
        Table of the WoS search result(s). A column indicating the source
        file name is added
    selected : dict
        Selected records
    """
    selected_studies = {
        int(k[7:]): '__'.join(b) for k, (a, b) in selected.items() if a}
    records_selection = record_pd.loc[sorted(selected_studies), :]

    if not len(records_selection):
        print('Nothing selected')
        sys.exit(0)

    records_selection['labels'] = [
        selected_studies[idx] for idx in records_selection.index]
    print('Written: %s' % output_fp)
    records_selection.to_csv(output_fp, index=False, sep='\t')
